# Copyright (C) Simula Research Laboratory and Jørgen S. Dokken, Cécile Daversin-Catty
# SPDX-License-Identifier:    MIT
"""Solver interface for graphs."""

import typing

from petsc4py import PETSc

import dolfinx.fem.petsc
import dolfinx.la.petsc
from networks_fenicsx import assembly

__all__ = ["Solver"]


class Solver:
    """PETSc solver interface for the Network problems.

    Args:
        assembler: The hydraulic network assembler.
        petsc_options_prefix: Prefix for PETSc options.
        petsc_options: Dictionary of PETSc options.
        kind: Kind of PETSc matrix and vectors to create.
    """

    _ksp: PETSc.KSP | None = None  # type: ignore[name-defined]
    _A: PETSc.Mat | None = None  # type: ignore[name-defined]
    _b: PETSc.Vec | None = None  # type: ignore[name-defined]
    _x: PETSc.Vec | None = None  # type: ignore[name-defined]
    _assembler: assembly.HydraulicNetworkAssembler

    def __init__(
        self,
        assembler: assembly.HydraulicNetworkAssembler,
        petsc_options_prefix: str = "NetworkSolver_",
        petsc_options: dict | None = None,
        kind: str | typing.Sequence[typing.Sequence[str]] | None = None,
    ):
        self._assembler = assembler

        self._ksp = PETSc.KSP().create(self._assembler.network.comm)  # type: ignore[attr-defined]

        self._A = dolfinx.fem.petsc.create_matrix(self._assembler.bilinear_forms, kind=kind)
        kind = "nest" if self._A.getType() == "nest" else kind  # type: ignore[attr-defined]
        assert isinstance(kind, str) or kind is None
        self._b = dolfinx.fem.petsc.create_vector(
            dolfinx.fem.extract_function_spaces(self._assembler.linear_forms), kind=kind
        )
        self._x = dolfinx.fem.petsc.create_vector(self._assembler.function_spaces, kind)

        self.ksp.setOperators(self.A)

        # Set PETSc options
        self.ksp.setOptionsPrefix(petsc_options_prefix)
        self._A.setOptionsPrefix(f"{petsc_options_prefix}A_")
        self._b.setOptionsPrefix(f"{petsc_options_prefix}b_")

        if petsc_options is None:
            petsc_options = {
                "ksp_type": "preonly",
                "pc_type": "lu",
                "pc_factor_mat_solver_type": "mumps",
                "ksp_monitor": None,
                "ksp_error_if_not_converged": True,
            }
        opts = PETSc.Options()  # type: ignore[attr-defined]
        opts.prefixPush(self.ksp.getOptionsPrefix())
        for key, value in petsc_options.items():
            opts[key] = value
        self.ksp.setFromOptions()
        self._A.setFromOptions()
        self._b.setFromOptions()
        opts.prefixPop()

    @property
    def assembler(self) -> assembly.HydraulicNetworkAssembler:
        """The hydraulic network assembler."""
        return self._assembler

    @property
    def A(self) -> PETSc.Mat:  # type: ignore[name-defined]
        """System matrix."""
        return self._A

    @property
    def b(self) -> PETSc.Vec:  # type: ignore[name-defined]
        """Right-hand side vector."""
        return self._b

    def assemble(self, lhs: bool = True, rhs: bool = True):
        """Assemble the system matrix and rhs vector.

        Args:
            lhs: Whether to assemble the system matrix.
            rhs: Whether to assemble the rhs vector.
        """
        if self._A is not None and lhs:
            self._A.zeroEntries()
        if self._b is not None and rhs:
            dolfinx.la.petsc._zero_vector(self._b)
        self.assembler.assemble(self._A, self._b, assemble_lhs=lhs, assemble_rhs=rhs)

    @property
    def ksp(self) -> PETSc.KSP:  # type: ignore[name-defined]
        return self._ksp

    @dolfinx.common.timed("nxfx:Solver:solve")
    def solve(
        self, functions: list[dolfinx.fem.Function] | None = None
    ) -> list[dolfinx.fem.Function]:
        """Solve the linear system of equations and assign them to a set of corresponding
        DOLFINx functions.

        Args:
            functions: List of DOLFINx functions to assign the solution to.
                If not provided they are created based on the assembler information.
        Returns:
            The functions.
        """
        if functions is None:
            functions = []
            for i, Vi in enumerate(self.assembler.flux_spaces):
                functions.append(dolfinx.fem.Function(Vi, name=f"flux_color_{i}"))
            functions.append(dolfinx.fem.Function(self.assembler.pressure_space, name="pressure"))
            functions.append(dolfinx.fem.Function(self.assembler.lm_space, name="global_flux"))

        self.ksp.solve(self.b, self._x)
        dolfinx.la.petsc._ghost_update(
            self._x,
            insert_mode=PETSc.InsertMode.INSERT,  # type: ignore[attr-defined]
            scatter_mode=PETSc.ScatterMode.FORWARD,  # type: ignore[attr-defined]
        )
        assert isinstance(self._x, PETSc.Vec)  # type: ignore[attr-defined]
        dolfinx.fem.petsc.assign(self._x, functions)
        return functions

    def __del__(self):
        if self._ksp is not None:
            self._ksp.destroy()
        if self._b is not None:
            self._b.destroy()
        if self._x is not None:
            self._x.destroy()



import pytest

from networks_fenicsx import NetworkMesh, network_generation


@pytest.mark.parametrize("gdim", [2, 3])
@pytest.mark.parametrize("N", [1, 4, 10])
@pytest.mark.parametrize("n", [2, 5, 7])
@pytest.mark.parametrize("H", [1, 2])
def test_make_tree(n: int, H: int, gdim: int, N: int):
    G = network_generation.make_tree(n=n, H=H, W=1, dim=gdim)
    network_mesh = NetworkMesh(G, N=N)

    domain = network_mesh.mesh

    tdim = domain.topology.dim
    assert tdim == 1
    assert domain.geometry.dim == gdim
    num_cells_global = domain.topology.index_map(tdim).size_global

    num_segments = sum(2**i for i in range(n))
    assert num_cells_global == N * num_segments
    num_vertices_global = domain.topology.index_map(0).size_global
    assert num_vertices_global == N + 1 + (num_segments - 1) * N



"""Top-level package for 1D networks simulations with FEniCSx."""

from importlib.metadata import metadata

meta = metadata("networks_fenicsx")
__version__ = meta["Version"]
__author__ = meta.get("Author", "")
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]

import networks_fenicsx.network_generation as network_generation
import networks_fenicsx.post_processing as post_processing

from .assembly import HydraulicNetworkAssembler
from .mesh import NetworkMesh
from .solver import Solver

__all__ = [
    "HydraulicNetworkAssembler",
    "NetworkMesh",
    "post_processing",
    "Solver",
    "network_generation",
]



# Copyright (C) Simula Research Laboratory, Cécile Daversin-Catty, Joe P. Dean and Jørgen S. Dokken
# SPDX-License-Identifier:    MIT
"""Assembly routine for Hydraulic network."""

import logging
import typing
from typing import Protocol

from petsc4py import PETSc

import numpy as np
import numpy.typing as npt

import basix
import dolfinx.la.petsc as _petsc_la
import ufl
from dolfinx import common, fem

from .mesh import NetworkMesh

__all__ = ["HydraulicNetworkAssembler", "PressureFunction"]


class PressureFunction(Protocol):
    def eval(self, x: npt.NDArray[np.floating]) -> npt.NDArray[np.inexact]: ...


@common.timed("nxfx:compute_integration_data")
def compute_integration_data(
    network_mesh: NetworkMesh,
) -> tuple[dict[int, npt.NDArray[np.int32]], dict[int, npt.NDArray[np.int32]]]:
    """Given a network mesh, compute integration entities for the "parent" network mesh
    for each bifuraction in the mesh.

    Args:
        network_mesh: The network mesh

    Returns:
        A tuple `(in_entities, out_entities) mapping integration entities
        on each edge of the network (marked by color) to its integration
        entities on the parent mesh.
    """

    # Pack all bifurcation in and out fluxes per colored edge in graph
    influx_color_to_bifurcations: dict[int, list[int]] = {
        int(color): [] for color in range(network_mesh._num_edge_colors)
    }
    outflux_color_to_bifurcations: dict[int, list[int]] = {
        int(color): [] for color in range(network_mesh._num_edge_colors)
    }
    for bifurcation in network_mesh.bifurcation_values:
        for color in network_mesh.in_edges(bifurcation):
            influx_color_to_bifurcations[color].append(int(bifurcation))
        for color in network_mesh.out_edges(bifurcation):
            outflux_color_to_bifurcations[color].append(int(bifurcation))
    # Accumulate integration data for all in-edges on the same submesh.
    in_flux_entities: dict[int, npt.NDArray[np.int32]] = {}
    out_flux_entities: dict[int, npt.NDArray[np.int32]] = {}

    for color in range(network_mesh._num_edge_colors):
        sm = network_mesh.submeshes[color]
        smfm = network_mesh.submesh_facet_markers[color]
        sm.topology.create_connectivity(sm.topology.dim - 1, sm.topology.dim)

        # Compute influx entities
        submesh_influx_entities = fem.compute_integration_domains(
            fem.IntegralType.exterior_facet,
            sm.topology,
            smfm.indices[np.isin(smfm.values, influx_color_to_bifurcations[color])],
        ).reshape(-1, 2)
        parent_to_sub_influx = network_mesh.entity_maps[color].sub_topology_to_topology(
            submesh_influx_entities[:, 0].copy(), inverse=False
        )
        submesh_influx_entities[:, 0] = parent_to_sub_influx
        in_flux_entities[color] = submesh_influx_entities.flatten()

        # Compute influx entities
        submesh_outflux_entities = fem.compute_integration_domains(
            fem.IntegralType.exterior_facet,
            sm.topology,
            smfm.indices[np.isin(smfm.values, outflux_color_to_bifurcations[color])],
        ).reshape(-1, 2)
        parent_to_sub_outflux = network_mesh.entity_maps[color].sub_topology_to_topology(
            submesh_outflux_entities[:, 0].copy(), inverse=False
        )
        submesh_outflux_entities[:, 0] = parent_to_sub_outflux
        out_flux_entities[color] = submesh_outflux_entities.flatten()
    return in_flux_entities, out_flux_entities


class HydraulicNetworkAssembler:
    """
    Assembler for the variational formulation of an hydraulic network.

    .. math::
        R q + \\frac{\\mathrm{d}}{\\mathrm{d}s} p = 0 \\\\
        \\frac{\\mathrm{d}}{\\mathrm{d}s} q  = f

    Args:
        mesh: The network mesh
        flux_degree: The polynomial degree for the flux functions
        pressure_degree: The polynomial degree for the pressure functions
    """

    _network_mesh: NetworkMesh
    _flux_spaces: list[fem.FunctionSpace]
    _pressure_space: fem.FunctionSpace
    _lm_space: fem.FunctionSpace
    _in_idx: int  # Starting point for each influx interior bifurcation integral
    _out_idx: int  # Starting point for each outflux interior bifurcation integral
    _in_keys: tuple[int, ...]  # Set of unique markers for all influx conditions
    _out_keys: tuple[int, ...]  # Set of unique markers for all outflux conditions
    _a: list[list[fem.Form]]  # Bilinear forms
    _L: list[fem.Form]  # Linear forms

    @common.timed("nxfx:HydraulicNetworkAssembler:__init__")
    def __init__(self, mesh: NetworkMesh, flux_degree: int = 1, pressure_degree: int = 0):
        self._network_mesh = mesh
        submeshes = self._network_mesh.submeshes

        # Flux spaces on each segment, ordered by the edge list
        # Using equispaced elements to match with legacy FEniCS
        flux_element = basix.ufl.element(
            family="Lagrange",
            cell="interval",
            degree=flux_degree,
            lagrange_variant=basix.LagrangeVariant.equispaced,
        )
        Pqs = [fem.functionspace(submsh, flux_element) for submsh in submeshes]

        if pressure_degree == 0:
            discontinuous = True
        else:
            discontinuous = False
        pressure_element = basix.ufl.element(
            family="Lagrange",
            cell="interval",
            degree=pressure_degree,
            lagrange_variant=basix.LagrangeVariant.equispaced,
            discontinuous=discontinuous,
        )
        Pp = fem.functionspace(self._network_mesh.mesh, pressure_element)

        self._flux_spaces = Pqs
        self._pressure_space = Pp
        self._lm_space = fem.functionspace(self._network_mesh.lm_mesh, ("DG", 0))

        # Compute integration data for network mesh
        self._integration_data = []
        self._in_idx = max(mesh.in_marker, mesh.out_marker) + 1
        in_flux_entities, out_flux_entities = compute_integration_data(self._network_mesh)
        self._in_keys = tuple(in_flux_entities.keys())
        self._out_keys = tuple(out_flux_entities.keys())
        for color in self._in_keys:
            self._integration_data.append((self._in_idx + color, in_flux_entities[color]))
        self._out_idx = self._in_idx + len(out_flux_entities)
        for color in self._out_keys:
            self._integration_data.append((self._out_idx + color, out_flux_entities[color]))

    @common.timed("nxfx:HydraulicNetworkAssembler:compute_forms")
    def compute_forms(
        self,
        p_bc_ex: typing.Callable[[npt.NDArray[np.floating]], npt.NDArray[np.inexact]]
        | fem.Expression
        | ufl.core.expr.Expr,
        f: ufl.core.expr.Expr | None = None,
        R: ufl.core.expr.Expr | None = None,
        jit_options: dict | None = None,
        form_compiler_options: dict | None = None,
    ):
        """
        Compute forms for hydraulic network model

        .. math::
            R q + \\frac{\\mathrm{d}}{\\mathrm{d}s} p = 0\\\\
            \\frac{\\mathrm{d}}{\\mathrm{d}s} q  = f


        on graph G, with bifurcation condition q_in = q_out
        and jump vectors the bifurcation conditions

        Args:
           f: source term
           p_bc: neumann bc for pressure
        """
        num_qs = self._network_mesh._num_edge_colors

        test_functions = [ufl.TestFunction(fs) for fs in self.function_spaces]
        trial_functions = [ufl.TrialFunction(fs) for fs in self.function_spaces]
        a: list[list[ufl.Form | ufl.ZeroBaseForm | None]] = [
            [ufl.ZeroBaseForm((ui, vj)) for vj in test_functions] for ui in trial_functions
        ]
        L: list[ufl.Form | ufl.ZeroBaseForm | None] = [
            ufl.ZeroBaseForm((vj,)) for vj in test_functions
        ]

        if f is None:
            f = fem.Constant(self._network_mesh.mesh, 0.0)

        if R is None:
            R = fem.Constant(self._network_mesh.mesh, 1.0)

        # Fluxes on each branch
        qs = []
        vs = []
        for Pq in self._flux_spaces:
            qs.append(ufl.TrialFunction(Pq))
            vs.append(ufl.TestFunction(Pq))

        # Lagrange multipliers
        lmbda = ufl.TrialFunction(self.lm_space)
        mu = ufl.TestFunction(self.lm_space)

        # Pressure
        p = ufl.TrialFunction(self._pressure_space)
        phi = ufl.TestFunction(self._pressure_space)
        # Assemble variational formulation
        network_mesh = self._network_mesh

        # Assemble edge contributions to a and L
        num_qs = len(self._network_mesh.submeshes)
        P1_e = fem.functionspace(network_mesh.mesh, ("Lagrange", 1))
        p_bc = fem.Function(P1_e)
        if isinstance(p_bc_ex, ufl.core.expr.Expr):
            try:
                expr = fem.Expression(p_bc_ex, P1_e.element.interpolation_points())  # type: ignore[operator]
            except TypeError:
                expr = fem.Expression(p_bc_ex, P1_e.element.interpolation_points)
            p_bc.interpolate(expr)
        else:
            p_bc.interpolate(p_bc_ex)

        dx_global = ufl.Measure("dx", domain=network_mesh.mesh)

        tangent = self._network_mesh.tangent
        for i, (submesh, entity_map, facet_marker) in enumerate(
            zip(
                network_mesh.submeshes,
                network_mesh.entity_maps,
                network_mesh.submesh_facet_markers,
            )
        ):
            dx_edge = ufl.Measure("dx", domain=submesh)
            ds_edge = ufl.Measure("ds", domain=submesh, subdomain_data=facet_marker)

            a[i][i] += R * qs[i] * vs[i] * dx_edge
            a[num_qs][i] += phi * ufl.dot(ufl.grad(qs[i]), tangent) * dx_edge
            a[i][num_qs] = -p * ufl.dot(ufl.grad(vs[i]), tangent) * dx_edge

            # Add all boundary contributions
            L[i] = p_bc * vs[i] * ds_edge(network_mesh.in_marker) - p_bc * vs[i] * ds_edge(
                network_mesh.out_marker
            )

        L[num_qs] += f * phi * dx_global

        # Multiplier mesh and flux share common parent mesh.
        # We create unique integration entities for each in and out branch
        # with a common ufl Measure
        # Map (bifurcation, edge) to local integration measure index
        ds = ufl.Measure(
            "ds", domain=self._network_mesh.mesh, subdomain_data=self._integration_data
        )
        for color in self._in_keys:
            a[-1][color] += mu * qs[color] * ds(self._in_idx + color)
            a[color][-1] += lmbda * vs[color] * ds(self._in_idx + color)

        for color in self._out_keys:
            a[-1][color] -= mu * qs[color] * ds(self._out_idx + color)
            a[color][-1] -= lmbda * vs[color] * ds(self._out_idx + color)

        entity_maps = [entity_map, self._network_mesh.lm_map, *self._network_mesh.entity_maps]

        # Replace remaining ZeroBaseForm with None
        # This is because `dolfinx.fem.forms._zero_form`
        # assumes single domain (no entity maps)
        for i, ai in enumerate(a):
            for j, aij in enumerate(ai):
                if isinstance(aij, ufl.ZeroBaseForm):
                    a[i][j] = None
        self._a = fem.form(
            a,  # type: ignore[arg-type]
            jit_options=jit_options,
            form_compiler_options=form_compiler_options,
            entity_maps=entity_maps,
        )
        self._L = fem.form(
            L,  # type: ignore[arg-type]
            jit_options=jit_options,
            form_compiler_options=form_compiler_options,
            entity_maps=entity_maps,
        )

    @property
    def lm_space(self) -> fem.FunctionSpace:
        """The function space of the bifurcation Lagrange multipliers"""
        return self._lm_space

    @property
    def pressure_space(self) -> fem.FunctionSpace:
        """The function space of the pressure function"""
        return self._pressure_space

    @property
    def flux_spaces(self) -> list[fem.FunctionSpace]:
        """List of function spaces for each flux function. The ith function corresponds
        to the edges of the :py:class:`networkx.DiGraph` that were colored with color `i`."""
        return self._flux_spaces

    @property
    def function_spaces(self) -> list[fem.FunctionSpace]:
        """List of all function-spaces in the order `[flux, pressure, lm]`,
        the same order as used by the :py:meth:`assemble`"""
        return [*self._flux_spaces, self._pressure_space, self._lm_space]

    @property
    def network(self) -> NetworkMesh:
        """Return the underlying network mesh."""
        return self._network_mesh

    @common.timed("nxfx:HydraulicNetworkAssembler:assemble")
    def assemble(
        self,
        A: PETSc.Mat | None = None,  # type: ignore[name-defined]
        b: PETSc.Mat | None = None,  # type: ignore[name-defined]
        assemble_lhs: bool = True,
        assemble_rhs: bool = True,
        kind: str | typing.Sequence[typing.Sequence[str]] | None = None,
    ) -> tuple[PETSc.Mat, PETSc.Vec]:  # type: ignore[name-defined]
        """Assemble system matrix and rhs vector.

        Note:
            If neither `A` or `b` is provided, they are created inside this class.

        Args:
            A: :py:class:`PETSc matrix<petsc4py.PETSc.Mat>` to assemble
                :py:attr:`HydraulicNetworkAssembler.bilinear_forms` into.
            b: :py:class:`PETSc vector<petsc4py.PETSc.Vec>` to assemble
                :py:attr:`HydraulicNetworkAssembler.linear_forms` into.
            assemble_lhs: Whether to assemble the system matrix.
            assemble_rhs: Whether to assemble the rhs vector.
            kind: If no matrix or vector is provided, "kind" is used to determine what
                kind of matrix/vector to create.
        """
        if assemble_lhs:
            if A is None:
                A = fem.petsc.create_matrix([[aij for aij in ai] for ai in self._a], kind=kind)
            A = fem.petsc.assemble_matrix(A, self._a, bcs=[])  # type: ignore
            A.assemble()
            kind = "nest" if A.getType() == PETSc.Mat.Type.NEST else kind  # type: ignore[attr-defined]
        if assemble_rhs:
            if b is None:
                assert isinstance(kind, str) or kind is None
                b = fem.petsc.create_vector(fem.extract_function_spaces(self._L), kind=kind)
            b = fem.petsc.assemble_vector(b, self._L)  # type: ignore
            _petsc_la._ghost_update(
                b,
                insert_mode=PETSc.InsertMode.ADD_VALUES,  # type: ignore[attr-defined]
                scatter_mode=PETSc.ScatterMode.REVERSE,  # type: ignore[attr-defined]
            )
        return (A, b)

    @property
    def bilinear_forms(self) -> typing.Sequence[typing.Sequence[fem.Form]]:
        """Nested list of the compiled, bilinear forms."""
        if self._a is None:
            logging.error("Bilinear forms haven't been computed. Need to call compute_forms()")
        else:
            return self._a

    def bilinear_form(self, i: int, j: int) -> fem.Form:
        """Extract the i,j bilinear form."""
        a = self.bilinear_forms
        if i > len(a) or j > len(a[i]):
            logging.error("Bilinear form a[" + str(i) + "][" + str(j) + "] out of range")
        return a[i][j]

    @property
    def linear_forms(self) -> typing.Sequence[fem.Form]:
        """Extract the linear form."""
        if self._L is None:
            logging.error("Linear forms haven't been computed. Need to call compute_forms()")
        else:
            return self._L

    def linear_form(self, i: int) -> fem.Form:
        """Return the i-th block of the linear form"""
        L = self.linear_forms
        if i > len(L):
            logging.error("Linear form L[" + str(i) + "] out of range")
        return L[i]



# Copyright (C) Alexandra Vallet, Simula Research Laboratory, Cécile Daversin-Catty
# and Jørgen S. Dokken
# SPDX-License-Identifier:    MIT

"""Utilities for generating :py:class:`Networkx Directional graphs<networkx.DiGraph>`"""

__all__ = ["make_tree", "make_arterial_tree"]

from typing import Annotated, Callable, Literal

import networkx as nx
import numpy as np
import numpy.typing as npt

import dolfinx.common


def tree_edges(n, r):
    # helper function for trees
    # yields edges in rooted tree at 0 with n nodes and branching ratio r
    if n == 0:
        return
    # Root branch
    source = 0
    target = 1
    yield source, target
    # Other branches
    nodes = iter(range(1, n))
    parents = [next(nodes)]  # stack of max length r
    while parents:
        source = parents.pop(0)
        for i in range(r):
            try:
                target = next(nodes)
                parents.append(target)
                yield source, target
            except StopIteration:
                break


@dolfinx.common.timed("nxfx:make_tree")
def make_tree(n: int, H: float, W: float, dim=3) -> nx.DiGraph:
    """
    Generate a symmetric tree whose root is at the origin.

    Args:
        n : Number of generations of branches
        H : Height of tree
        W : Width of tree at its largest extent
    """
    assert n >= 1, "Number of generations must be at least 1"
    r = 2  # FIXME : add parameter r : branching factor of the tree (each node has r children)
    G = nx.DiGraph()

    nb_nodes_gen = []
    for i in range(n):
        nb_nodes_gen.append(pow(r, i))

    nb_nodes = 1 + sum(nb_nodes_gen)
    nb_nodes_last = pow(r, n - 1)

    G.add_nodes_from(range(nb_nodes))

    x_offset = W / (2 * (nb_nodes_last - 1))
    y_offset = H / n

    # Add two first nodes
    idx = 0
    if dim == 2:
        G.nodes[idx]["pos"] = [0, 0]
        G.nodes[idx + 1]["pos"] = [0, y_offset]
    else:
        G.nodes[idx]["pos"] = [0, 0, 0]
        G.nodes[idx + 1]["pos"] = [0, y_offset, 0]
    idx = idx + 2

    # Add nodes for rest of the tree
    for gen in range(1, n):
        factor = pow(2, n - gen)
        x = x_offset * (factor / 2)
        y = y_offset * (gen + 1)
        x_coord = []
        nb_nodes_ = int(nb_nodes_gen[gen] / 2)
        for i in range(nb_nodes_):
            x_coord.append(x)
            x_coord.append(-x)
            x = x + x_offset * factor
        # Add nodes to G, from sorted x_coord array
        x_coord.sort()
        for x in x_coord:
            if dim == 2:
                G.nodes[idx]["pos"] = [x, y]
            else:
                G.nodes[idx]["pos"] = [x, y, 0]
            idx = idx + 1

    edges = tree_edges(nb_nodes, r)
    for e0, e1 in list(edges):
        G.add_edge(e0, e1)
    return G


def _default_normal(x: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
    """Surface plane normal orientation for the xy-plane"""
    output = np.zeros_like(x)
    output[2] = 1
    return output


def _project_onto_plane(x, n):
    """Project `x` onto plane with normal `n`"""
    d = np.dot(x, n) / np.linalg.norm(n)
    p = d * n / np.linalg.norm(n)
    return x - p


def _rotate_in_plane(
    x: Annotated[npt.NDArray[np.floating], Literal[(3,)]],
    axis: Annotated[npt.NDArray[np.floating], Literal[(3,)]],
    angle: float,
) -> npt.NDArray[np.floating]:
    """Use Rodrigues formula to rotate a vector in space given an axis and
    angle of rotation. Ref: https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    """
    theta = np.radians(angle)
    k = axis / np.linalg.norm(axis)
    K = np.array([[0, -k[2], k[1]], [k[2], 0, -k[0]], [-k[1], k[0], 0]])
    R = np.eye(3) + np.sin(theta) * K + (1 - np.cos(theta)) * np.dot(K, K)
    return np.dot(R, x)


def _compute_vessel_endpoint(previousvessel, surfacenormal, angle, length):
    """From a previous vessel defined in 3D,
    the brain surface at the end of the previous vessel, angle and length :
    compute the coordinate of the end node of the current vessel"""
    # project the previous vessel in the current plane
    # and get the direction vector of the previous vessel
    pm1 = previousvessel[0]
    p0 = previousvessel[1]
    vector_previous = p0 - pm1
    previousdir = _project_onto_plane(vector_previous, surfacenormal)
    # compute the direction vector of the new vessel with the angle
    newdir = _rotate_in_plane(previousdir, surfacenormal, angle)
    # compute the location of the end of the new vessel
    return _translate(p0, newdir, length)


def _translate(
    p0: npt.NDArray[np.floating], direction: npt.NDArray[np.floating], length: float
) -> npt.NDArray[np.floating]:
    """Compute `p1 = p0 + length * direction/|direction|`"""
    # compute the location of the end of the new vessel
    assert len(p0) == len(direction)
    return p0 + length * direction / np.linalg.norm(direction, axis=-1)


@dolfinx.common.timed("nxfx:make_arterial_tree")
def make_arterial_tree(
    N: int,
    p0: Annotated[npt.NDArray[np.floating], Literal[(3,)]] = np.zeros(3, dtype=np.float64),
    direction: Annotated[npt.NDArray[np.floating], Literal[(3,)]] = np.array(
        [0, 1, 0], dtype=np.float64
    ),
    D0: float = 2.0,
    lmbda: float = 8.0,
    gamma: float = 0.8,
    normal: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]] = _default_normal,
    random: bool = False,
) -> nx.DiGraph:
    """Create an arterial tree.

    Murray's law
    We consider that the sum of power 3 of daughter vessels radii equal
    to the power three of the parent vessel radius :math:`D0^3=D1^3 + D2^3`
    We consider that the ratio between two daughter vessels radii is gam
    :math:`D2/D1=gam`.
    Then we have
    
    .. math::
        D2=D0(\\gammma^3+1)^(-1/3)\\\\
        D1=\\gamma D2

    We consider that the length `L` of a vessel segment in a given network is
    related to its diameter `d` via :math:`L = \\lambda d`,
    where the positive constant :math:`\\lambda` is network-specific.

    The angle of bifurcation can be expressed based on blood volume conservation
    and minimum energy hypothesis
    FUNG, Y. (1997). Biomechanics: Circulation. 2nd edition. Heidelberg: Springer.
    HUMPHREY, J. and DELANGE, S. (2004). An Introduction to Biomechanics. Heidelberg: Springer.
    VOGEL, S. (1992). Vital Circuits. Oxford: Oxford University Press.
    See here for derivation : https://www.montogue.com/blog/murrays-law-and-arterial-bifurcations/

    Note:
        Code based on the repository: https://gitlab.com/ValletAlexandra/NetworkGen
        by Alexandra Vallet under MIT License.
    
    Args:
        N: Number of generations of vessels
        p0: Origin location
        direction: Initial direction
        D0: First vessel diameter
        lmbda: Network specific constant relating vessel segment length to diameter `d`.
        gamma: Ratio between two daughter vessels.
        normal: Function that computes the normal to the plane at any point (x,y,z)
        random: If True set randomly which vessel goes right or left. If False,
            the biggest vessel is always the second one
    """
    if gamma > 1:
        raise ValueError("Please choose a gamma lower or equal to 1")

    # Create directed graph
    G = nx.DiGraph()

    G.add_edge(0, 1)

    # Set initial vessel
    nx.set_node_attributes(G, p0, "pos")
    nx.set_edge_attributes(G, D0 / 2, "radius")
    L = D0 * lmbda
    G.nodes[1]["pos"] = _translate(p0, direction, L)

    inode = 1

    # List of vessels from the previous generation
    previous_edges = [(0, 1)]
    previous_vessel = np.empty((2, 3), dtype=p0.dtype)
    for _ in range(1, N):
        current_edges = []
        for e in previous_edges:
            # Parent vessel properties
            previous_vessel[0, :] = G.nodes[e[0]]["pos"]
            previous_vessel[1, :] = G.nodes[e[1]]["pos"]
            D0 = G.edges[e]["radius"] * 2

            # Daughter diameters
            D2 = D0 * (gamma**3 + 1) ** (-1 / 3)
            D1 = gamma * D2
            # Daughter lenghts
            L2 = lmbda * D2
            L1 = lmbda * D1
            # Bifurcation angles
            # angle for the smallest vessel
            cos1 = (D0**4 + D1**4 - (D0**3 - D1**3) ** (4 / 3)) / (2 * D0**2 * D1**2)
            angle1 = np.degrees(np.arccos(cos1))
            # angle for the biggest vessel
            cos2 = (D0**4 + D2**4 - (D0**3 - D2**3) ** (4 / 3)) / (2 * D0**2 * D2**2)
            angle2 = np.degrees(np.arccos(cos2))
            sign1 = 1 if not random else np.random.choice([-1, 1])
            sign2 = -sign1

            # Add first daughter vessel
            inode += 1
            new_edge = (e[1], inode)
            G.add_edge(*new_edge)

            # Compute endpoint of vessel
            G.nodes[inode]["pos"] = _compute_vessel_endpoint(
                previous_vessel, normal(previous_vessel[1]), sign1 * angle1, L1
            )

            # Set radius
            G.edges[new_edge]["radius"] = D1 / 2

            # Add to the pool of vessels for this generation
            current_edges.append(new_edge)

            inode += 1
            new_edge = (e[1], inode)
            G.add_edge(*new_edge)

            # Set the location according to length and angle
            G.nodes[inode]["pos"] = _compute_vessel_endpoint(
                previous_vessel, normal(previous_vessel[1]), sign2 * angle2, L2
            )

            # Set radius
            G.edges[new_edge]["radius"] = D2 / 2

            # Add to the pool of vessels for this generation
            current_edges.append(new_edge)
        previous_edges = current_edges
    return G



# Copyright (C) Simula Research Laboratory and Jørgen S. Dokken
# SPDX-License-Identifier:    MIT
"""
Interface for converting a :py:class:`directional networkx graph<networkx.DiGraph>`
into a :py:class:`dolfinx.mesh.Mesh`.

This idea stems from the Graphnics project (https://doi.org/10.48550/arXiv.2212.02916),
https://github.com/IngeborgGjerde/fenics-networks by Ingeborg Gjerde.
"""

import inspect
from typing import Any, Callable, Iterable

from mpi4py import MPI

import networkx as nx
import numpy as np
import numpy.typing as npt

import basix.ufl
import ufl
from dolfinx import common, fem, mesh
from dolfinx import graph as _graph
from dolfinx import io as _io

__all__ = ["NetworkMesh"]


@common.timed("nxfx:color_graph")
def color_graph(
    graph: nx.DiGraph,
    strategy: str | Callable[[nx.Graph, dict[int, int]], Iterable[int]] | None,
) -> dict[tuple[int, int], int]:
    """
    Color the edges of a graph.
    """
    if strategy is not None:
        undirected_edge_graph = nx.line_graph(graph.to_undirected())
        edge_coloring = nx.coloring.greedy_color(undirected_edge_graph, strategy=strategy)
    else:
        edge_coloring = {edge: i for i, edge in enumerate(graph.edges)}
    return edge_coloring


class NetworkMesh:
    """A representation of a :py:class:`Directional Networkx<networkx.DiGraph>` graph in
    :py:mod:`DOLFINx<dolfinx>`.

    Stores the resulting :py:class:`mesh<dolfinx.mesh.Mesh>`,
    :py:class:`subdomains<dolfinx.mesh.MeshTags>`,
    and :py:class:`facet markers<dolfinx.mesh.MeshTags>` for bifurcations and boundary nodes.
    Has a :py:class:`submesh<dolfinx.mesh.Mesh>` for each edge in the
    :py:class:`Networkx graph<networkx.DiGraph>`.

    Args:
        graph: The directional networkx graph to convert.
        N: Number of elements per segment.
        color_strategy: Strategy to use for coloring the graph edges.
            If set to `None`, no-graphcoloring is used (not recommended).
        comm: The MPI communicator to distribute the mesh on.
        graph_rank: The MPI rank of the process that holds the graph.
    """

    # Graph properties
    _geom_dim: int
    _num_edge_colors: int
    _bifurcation_in_color: dict[int, list[int]]
    _bifurcation_out_color: dict[int, list[int]]

    # Mesh properties
    _msh: mesh.Mesh | None
    _subdomains: mesh.MeshTags | None
    _facet_markers: mesh.MeshTags | None
    _submesh_facet_markers: list[mesh.MeshTags]
    _edge_meshes: list[mesh.Mesh]
    _edge_entity_maps: list[mesh.EntityMap]
    _tangent: fem.Function
    _bifurcation_values: npt.NDArray[np.int32]
    _boundary_values: npt.NDArray[np.int32]
    _lm_mesh: mesh.Mesh | None
    _lm_map: mesh.EntityMap | None

    def __init__(
        self,
        graph: nx.DiGraph,
        N: int,
        color_strategy: str | Callable[[nx.Graph, dict[int, int]], Iterable[int]] | None = None,
        comm: MPI.Comm = MPI.COMM_WORLD,
        graph_rank: int = 0,
    ):
        self._build_mesh(
            graph, N=N, color_strategy=color_strategy, comm=comm, graph_rank=graph_rank
        )
        self._build_network_submeshes()
        self._create_lm_submesh()

    @property
    def lm_mesh(self) -> mesh.Mesh:
        """Lagrange multiplier mesh, a point-cloud mesh including each bifurcation."""
        if self._lm_mesh is None:
            raise RuntimeError("Lagrange multiplier submesh has not been created.")
        return self._lm_mesh

    @property
    def lm_map(self) -> mesh.EntityMap:
        """Entity map for the :py:meth:`Lagrange multiplier mesh<NetworkMesh.lm_mesh>`"""
        if self._lm_map is None:
            raise RuntimeError("Lagrange multiplier entity map has not been created.")
        return self._lm_map

    @property
    def comm(self) -> MPI.Comm:
        """MPI-communicator of the network mesh"""
        return self.mesh.comm

    @common.timed("nxfx:NetworkMesh:create_lm_submesh")
    def _create_lm_submesh(self):
        """Create a submesh for the Lagrange multipliers at the bifurcations.

        Note:
            This is an internal class function, that is not supposed to be called by
            users.
        """
        assert self._msh is not None
        assert self._facet_markers is not None
        bifurcation_indices = self._facet_markers.indices[
            np.isin(self._facet_markers.values, self.bifurcation_values)
        ]
        self._lm_mesh, self._lm_map = mesh.create_submesh(
            self.mesh,
            self.mesh.topology.dim - 1,
            bifurcation_indices,
        )[:2]
        # Workaround until: https://github.com/FEniCS/dolfinx/pull/3974 is merged
        self._lm_mesh.topology.create_entity_permutations()

    @common.timed("nxfx:NetworkMesh:build_mesh")
    def _build_mesh(
        self,
        graph: nx.DiGraph | None,
        N,
        color_strategy: str | Callable[[nx.Graph, dict[int, int]], Iterable[int]] | None,
        comm: MPI.Comm,
        graph_rank: int,
    ):
        """Convert the networkx graph into a {py:class}`dolfinx.mesh.Mesh`.

        Each segment in the networkx graph gets a unique subdomain marker.
        Each bifurcation and boundary node is marked on the facets with a unique integer.

        Note:
            This function attaches data to `self.mesh`, `self.subdomains` and
            `self.boundaries`.

        Args:
            graph: The networkx graph to convert.
            N: Number of elements per segment.
            color_strategy: Strategy to use for coloring the graph edges.
                If set to `None`, no-graphcoloring is used (not recommended).
            comm: The MPI communicator to distribute the mesh on.
            graph_rank: The MPI rank of the process that holds the graph.
        """
        # Extract geometric dimension from graph
        geom_dim = None
        num_edge_colors = None
        max_connections = None
        bifurcation_values = None
        boundary_values = None
        number_of_nodes = None
        bifurcation_in_color: dict[int, list[int]] | None = None
        bifurcation_out_color: dict[int, list[int]] | None = None
        if comm.rank == graph_rank:
            assert isinstance(graph, nx.DiGraph), f"Directional graph not present of {graph_rank}"
            geom_dim = len(graph.nodes[1]["pos"])
            edge_coloring = color_graph(graph, color_strategy)

            num_edge_colors = len(set(edge_coloring.values()))
            cells_array = np.asarray([[u, v] for u, v in graph.edges()])
            number_of_nodes = graph.number_of_nodes()
            nodes_with_degree = np.full(number_of_nodes, -1, dtype=np.int32)
            for node, degree in graph.degree():
                nodes_with_degree[node] = degree
            bifurcation_values = np.flatnonzero(nodes_with_degree > 1)
            boundary_values = np.flatnonzero(nodes_with_degree == 1)
            max_connections = np.max(nodes_with_degree)

            bifurcation_in_color = {}
            bifurcation_out_color = {}
            for bifurcation in bifurcation_values:
                in_edges = graph.in_edges(bifurcation)
                bifurcation_in_color[int(bifurcation)] = []
                for edge in in_edges:
                    bifurcation_in_color[int(bifurcation)].append(edge_coloring[edge])
                out_edges = graph.out_edges(bifurcation)
                bifurcation_out_color[int(bifurcation)] = []
                for edge in out_edges:
                    bifurcation_out_color[int(bifurcation)].append(edge_coloring[edge])

            # Map boundary_values to inlet and outlet data from graph
            boundary_in_nodes = []
            boundary_out_nodes = []
            for boundary in boundary_values:
                in_edges = graph.in_edges(boundary)
                out_edges = graph.out_edges(boundary)
                assert len(in_edges) + len(out_edges) == 1, "Boundary node with multiple edges"
                if len(in_edges) == 1:
                    boundary_in_nodes.append(int(boundary))
                else:
                    boundary_out_nodes.append(int(boundary))

        comm.bcast(num_edge_colors, root=graph_rank)
        num_edge_colors, number_of_nodes, max_connections, geom_dim = comm.bcast(
            (num_edge_colors, number_of_nodes, max_connections, geom_dim), root=graph_rank
        )
        comm.barrier()
        bifurcation_values, boundary_values = comm.bcast(
            (bifurcation_values, boundary_values), root=graph_rank
        )
        comm.barrier()
        bifurcation_in_color, bifurcation_out_color = comm.bcast(
            (bifurcation_in_color, bifurcation_out_color), root=graph_rank
        )
        comm.barrier()

        self._geom_dim = geom_dim
        self._num_edge_colors = num_edge_colors
        self._bifurcation_values = bifurcation_values
        self._boundary_values = boundary_values

        # Create lookup of in and out colors for each bifurcation
        self._bifurcation_in_color = bifurcation_in_color
        self._bifurcation_out_color = bifurcation_out_color

        # Generate mesh
        tangents: npt.NDArray[np.float64] = np.empty((0, self._geom_dim), dtype=np.float64)
        if comm.rank == graph_rank:
            assert isinstance(graph, nx.DiGraph), (
                f"No directional graph present on rank {comm.rank}"
            )
            vertex_coords = np.asarray([graph.nodes[v]["pos"] for v in graph.nodes()])
            line_weights = np.linspace(0, 1, N, endpoint=False)[1:][:, None]
            mesh_nodes = vertex_coords.copy()
            cells = []
            cell_markers = []
            local_tangents = []
            if len(line_weights) == 0:
                for segment in cells_array:
                    cells.append(np.array([segment[0], segment[1]], dtype=np.int64))
                    cell_markers.append(edge_coloring[(segment[0], segment[1])])
                    start = vertex_coords[segment[0]]
                    end = vertex_coords[segment[1]]
                    local_tangents.append(_compute_tangent(start, end))
            else:
                for segment in cells_array:
                    start_coord_pos = mesh_nodes.shape[0]
                    start = vertex_coords[segment[0]]
                    end = vertex_coords[segment[1]]
                    internal_line_coords = start * (1 - line_weights) + end * line_weights

                    tangent = _compute_tangent(start, end)
                    local_tangents.append(np.tile(tangent, (internal_line_coords.shape[0] + 1, 1)))

                    mesh_nodes = np.vstack((mesh_nodes, internal_line_coords))
                    cells.append(np.array([segment[0], start_coord_pos], dtype=np.int64))
                    segment_connectivity = (
                        np.repeat(np.arange(internal_line_coords.shape[0]), 2)[1:-1].reshape(
                            internal_line_coords.shape[0] - 1, 2
                        )
                        + start_coord_pos
                    )
                    cells.append(segment_connectivity)
                    cells.append(
                        np.array(
                            [
                                start_coord_pos + internal_line_coords.shape[0] - 1,
                                segment[1],
                            ],
                            dtype=np.int64,
                        )
                    )
                    cell_markers.extend(
                        np.full(
                            internal_line_coords.shape[0] + 1,
                            edge_coloring[(segment[0], segment[1])],
                            dtype=np.int32,
                        )
                    )
            cells_ = np.vstack(cells).astype(np.int64)
            cell_markers_ = np.array(cell_markers, dtype=np.int32)
            tangents = np.vstack(local_tangents).astype(np.float64)
        else:
            cells_ = np.empty((0, 2), dtype=np.int64)
            mesh_nodes = np.empty((0, self._geom_dim), dtype=np.float64)
            cell_markers_ = np.empty((0,), dtype=np.int32)
        tangents = tangents[:, : self._geom_dim].copy()
        sig = inspect.signature(mesh.create_cell_partitioner)
        part_kwargs = {}
        if "max_facet_to_cell_links" in list(sig.parameters.keys()):
            part_kwargs["max_facet_to_cell_links"] = np.max(max_connections)
        partitioner = mesh.create_cell_partitioner(mesh.GhostMode.shared_facet, **part_kwargs)
        sig = inspect.signature(mesh.create_mesh)
        kwargs = {}
        if "max_facet_to_cell_links" in list(sig.parameters.keys()):
            kwargs["max_facet_to_cell_links"] = np.max(max_connections)

        graph_mesh = mesh.create_mesh(
            comm,
            x=mesh_nodes,
            cells=cells_,
            e=ufl.Mesh(basix.ufl.element("Lagrange", "interval", 1, shape=(self._geom_dim,))),
            partitioner=partitioner,
            **kwargs,
        )
        self._msh = graph_mesh

        local_entities, local_values = _io.distribute_entity_data(
            self.mesh,
            self.mesh.topology.dim,
            cells_,
            cell_markers_,
        )

        # Distribute tangents
        tangent_space = fem.functionspace(graph_mesh, ("DG", 0, (self._geom_dim,)))
        self._tangent = fem.Function(tangent_space)

        # Which cells from the input mesh (rank 0) correspond to the local cells on this rank
        original_cell_indices = self.mesh.topology.original_cell_index[
            : self.mesh.topology.index_map(self.mesh.topology.dim).size_local
        ]

        # First compute how much data we are getting from each rank
        send_count = original_cell_indices.size
        recv_counts = comm.gather(send_count, root=graph_rank)
        if comm.rank == graph_rank:
            assert isinstance(recv_counts, Iterable)
            assert sum(recv_counts) == cells_.shape[0]

        # Send to rank 0 which tangents that we need
        recv_buffer: None | list[npt.NDArray[np.int64] | list[Any] | MPI.Datatype] = None
        if comm.rank == graph_rank:
            assert recv_counts is not None
            recv_buffer = [np.empty(cells_.shape[0], dtype=np.int64), recv_counts, MPI.INT64_T]

        comm.Gatherv(sendbuf=original_cell_indices, recvbuf=recv_buffer, root=graph_rank)
        send_t_buffer = None
        if comm.rank == graph_rank:
            assert recv_buffer is not None
            recv_data = recv_buffer[0]
            assert isinstance(recv_data, np.ndarray)
            send_t_buffer = [
                tangents[recv_data].flatten(),
                np.array(recv_counts) * self._geom_dim,
            ]

        comm.Scatterv(sendbuf=send_t_buffer, recvbuf=self.tangent.x.array, root=graph_rank)
        self.tangent.x.scatter_forward()

        self._subdomains = mesh.meshtags_from_entities(
            self.mesh,
            self.mesh.topology.dim,
            _graph.adjacencylist(local_entities),
            local_values,
        )

        self._in_marker = 3 * number_of_nodes
        self._out_marker = 5 * number_of_nodes
        if comm.rank == graph_rank:
            lv = np.arange(number_of_nodes, dtype=np.int64).reshape(-1, 1)
            lvv = np.arange(number_of_nodes, dtype=np.int32)
            lvv[boundary_in_nodes] = self._in_marker
            lvv[boundary_out_nodes] = self._out_marker
        else:
            lv = np.empty((0, 1), dtype=np.int64)
            lvv = np.empty((0,), dtype=np.int32)

        self.mesh.topology.create_connectivity(0, 1)
        local_vertices, local_vertex_values = _io.distribute_entity_data(self.mesh, 0, lv, lvv)
        self._facet_markers = mesh.meshtags_from_entities(
            self.mesh,
            0,
            _graph.adjacencylist(local_vertices),
            local_vertex_values,
        )

        self.subdomains.name = "subdomains"
        self.boundaries.name = "bifurcations"

    @common.timed("nxfx:NetworkMesh:build_network_submeshes")
    def _build_network_submeshes(self):
        """Create submeshes for each edge in the network."""
        assert self._msh is not None
        assert self._subdomains is not None
        assert self._facet_markers is not None
        self._edge_meshes = []
        self._edge_entity_maps = []
        self._submesh_facet_markers = []
        parent_vertex_map = self.mesh.topology.index_map(0)
        num_vertices_parent = parent_vertex_map.size_local + parent_vertex_map.num_ghosts
        parent_vertex_marker = np.full(num_vertices_parent, -1, dtype=np.int32)
        parent_vertex_marker[self._facet_markers.indices] = self._facet_markers.values
        for i in range(self._num_edge_colors):
            # Create submesh of color i
            edge_subdomain = self.subdomains.indices[self.subdomains.values == i]
            edge_mesh, edge_map, vertex_map = mesh.create_submesh(
                self.mesh, self.mesh.topology.dim, edge_subdomain
            )[0:3]
            self._edge_meshes.append(edge_mesh)
            self._edge_entity_maps.append(edge_map)

            # Map all submesh vertices to parent
            num_submesh_vertices = (
                edge_mesh.topology.index_map(0).size_local
                + edge_mesh.topology.index_map(0).num_ghosts
            )
            parent_vertices = vertex_map.sub_topology_to_topology(
                np.arange(num_submesh_vertices, dtype=np.int32), inverse=False
            )
            sub_topology_values = parent_vertex_marker[parent_vertices]
            marked_vertices = np.flatnonzero(sub_topology_values >= 0)
            marked_values = sub_topology_values[marked_vertices].copy()
            self._submesh_facet_markers.append(
                mesh.meshtags(edge_mesh, 0, marked_vertices, marked_values)
            )

    def in_edges(self, bifurcation: int) -> list[int]:
        return self._bifurcation_in_color[bifurcation]

    def out_edges(self, bifurcation: int) -> list[int]:
        return self._bifurcation_out_color[bifurcation]

    @property
    def submesh_facet_markers(self) -> list[mesh.MeshTags]:
        if self._submesh_facet_markers is None:
            raise RuntimeError("Mesh has no submesh facet markers")
        return self._submesh_facet_markers

    @property
    def mesh(self):
        if self._msh is None:
            raise RuntimeError("Mesh has not been built yet. Call build_mesh() first.")
        return self._msh

    @property
    def subdomains(self):
        if self._subdomains is None:
            raise RuntimeError("Mesh has no subdomains")
        return self._subdomains

    @property
    def boundaries(self):
        if self._facet_markers is None:
            raise RuntimeError("Mesh has no boundaries/facet markers")
        return self._facet_markers

    @property
    def submeshes(self):
        if len(self._edge_meshes) == 0:
            raise RuntimeError(
                "Submeshes have not been built yet. Call build_network_submeshes() first."
            )
        return self._edge_meshes

    @property
    def entity_maps(self):
        if len(self._edge_entity_maps) == 0:
            raise RuntimeError(
                "Entity maps have not been built yet. Call build_network_submeshes() first."
            )
        return self._edge_entity_maps

    @property
    def tangent(self):
        """Return DG-0 field containing the tangent vector of the graph."""
        return self._tangent

    @property
    def bifurcation_values(self) -> npt.NDArray[np.int32]:
        return self._bifurcation_values

    @property
    def boundary_values(self) -> npt.NDArray[np.int32]:
        return self._boundary_values

    @property
    def in_marker(self) -> int:
        return self._in_marker

    @property
    def out_marker(self) -> int:
        return self._out_marker


def _compute_tangent(
    start: npt.NDArray[np.float64], end: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """Compute the tangent vector from start to end points.

    Tangent is oriented according to positive y-axis.
    If perpendicular to y-axis, align with x-axis.

    Args:
        start: Starting point coordinates.
        end: Ending point coordinates.

    Returns:
        Normalized tangent vector from start to end.
    """
    tangent = end - start
    tangent /= np.linalg.norm(tangent, axis=-1)
    return tangent



# Copyright (C) Simula Research Laboratory and Jørgen S. Dokken
# SPDX-License-Identifier:    MIT
"""
Convenience functions for post-processing.
"""

from pathlib import Path

from mpi4py import MPI

import numpy as np

from dolfinx import fem, io
from networks_fenicsx.mesh import NetworkMesh

__all__ = ["extract_global_flux", "export_functions", "export_submeshes"]


def extract_global_flux(graph_mesh: NetworkMesh, functions: list[fem.Function]) -> fem.Function:
    """Extract global flux function from submesh solutions

    Args:
        graph_mesh: The network mesh
        functions: The list of functions `[flux_1, flux_2, ..., flux_M, pressure, lm]`
    """
    flux_functions = functions[:-2]

    # Extract info regarding global flux space
    q_space = flux_functions[0].function_space
    q_degree = q_space.element.basix_element.degree
    global_q_space = fem.functionspace(graph_mesh.mesh, ("DG", q_degree))
    global_q = fem.Function(global_q_space, name="Global_Flux")

    # Recover solution
    # Flux spaces are the first M ones
    for i, (flux, entity_map) in enumerate(zip(flux_functions, graph_mesh.entity_maps)):
        # Interpolated to DG space
        flux.name = f"Flux_{i}"
        submesh = flux.function_space.mesh
        num_cells_local = (
            submesh.topology.index_map(submesh.topology.dim).size_local
            + submesh.topology.index_map(submesh.topology.dim).num_ghosts
        )
        submesh_to_parent = entity_map.sub_topology_to_topology(
            np.arange(num_cells_local, dtype=np.int32), inverse=False
        )
        global_q.interpolate(
            flux,
            cells0=np.arange(num_cells_local, dtype=np.int32),
            cells1=submesh_to_parent,
        )
    return global_q


def export_functions(
    functions: list[fem.Function],
    outpath: Path | str,
):
    """Export global flux function from submesh solutions.
    Args:
        functions: The list of functions `[flux_1, flux_2, ..., flux_M, pressure, lm]`
        outpath: The output directory to save the files to.

    """
    export_path = Path(outpath)
    flux_functions = functions[:-2]
    for i, q in enumerate(flux_functions):
        with io.VTXWriter(
            MPI.COMM_WORLD,
            export_path / f"flux_{i}.bp",
            q,
        ) as f:
            f.write(0.0)

    with io.VTXWriter(MPI.COMM_WORLD, export_path / "pressure.bp", functions[-2]) as f:
        f.write(0.0)
    with io.VTXWriter(MPI.COMM_WORLD, export_path / "lm.bp", functions[-1]) as f:
        f.write(0.0)


def export_submeshes(network_mesh: NetworkMesh, outpath: str | Path):
    """Export syvmeshes with appropriate facet markers to file.

    Args:
        network_mesh: The network mesh
        outpath: Folder to place the submeshes in
    """
    outpath = Path(outpath)
    for i in range(network_mesh._num_edge_colors):
        network_mesh.submeshes[i].topology.create_connectivity(0, 1)
        with io.XDMFFile(
            network_mesh.submeshes[i].comm, outpath / f"submesh_{i}.xdmf", "w"
        ) as xdmf:
            xdmf.write_mesh(network_mesh.submeshes[i])
            xdmf.write_meshtags(
                network_mesh.submesh_facet_markers[i], network_mesh.submeshes[i].geometry
            )
